use nalgebra::Vector3 as Vec3;
use serde::{Deserialize, Serialize};
use slotmap::SlotMap;
use std::collections::HashMap;

use crate::id_types::*;
use crate::SVec;

#[derive(Debug, Clone, Copy, PartialEq, Eq, Serialize, Deserialize)]
pub enum DataType {
    Vector,
    Scalar,
    Selection,
    Mesh,
    Enum,
    // The path to a (possibly new) file where export contents will be saved to
    NewFile,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub enum InputParamMetadata {
    MinMaxScalar { min: f32, max: f32 },
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub enum InputParamValue {
    Vector(Vec3<f32>),
    Scalar(f32),
    Selection {
        text: String,
        selection: Option<Vec<u32>>,
    },
    /// Used for parameters that can't have a value because they only accept
    /// connections.
    None,
    Enum {
        values: Vec<String>,
        selection: Option<u32>,
    },
    NewFile {
        path: Option<std::path::PathBuf>,
    },
}

/// There are three kinds of input params
#[derive(Debug, Clone, Copy, Serialize, Deserialize)]
pub enum InputParamKind {
    /// No constant value can be set. Only incoming connections can produce it
    ConnectionOnly,
    /// Only a constant value can be set. No incoming connections accepted.
    ConstantOnly,
    /// Both incoming connections and constants are accepted. Connections take
    /// precedence over the constant values.
    ConnectionOrConstant,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct InputParam {
    pub id: InputId,
    /// The data type of this node. Used to determine incoming connections. This
    /// should always match the type of the InputParamValue, but the property is
    /// not actually enforced.
    pub typ: DataType,
    /// The constant value stored in this parameter.
    pub value: InputParamValue,
    /// A list of metadata fields, specifying things like bounds or limits.
    /// Metadata values that don't make sense for a type are ignored.
    pub metadata: SVec<InputParamMetadata>,
    /// The input kind. See [InputParamKind]
    pub kind: InputParamKind,
    /// Back-reference to the node containing this parameter.
    pub node: NodeId,
}

impl InputParam {
    pub fn value(&self) -> InputParamValue {
        self.value.clone()
    }

    pub fn kind(&self) -> InputParamKind {
        self.kind
    }

    pub fn node(&self) -> NodeId {
        self.node
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct OutputParam {
    pub id: OutputId,
    /// Back-reference to the node containing this parameter.
    pub node: NodeId,
    pub typ: DataType,
}

impl OutputParam {
    pub fn node(&self) -> NodeId {
        self.node
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Node {
    pub id: NodeId,
    pub label: String,
    pub op_name: String,
    pub inputs: Vec<(String, InputId)>,
    pub outputs: Vec<(String, OutputId)>,
    /// Executable nodes will run some code when their "Run" button is clicked
    pub is_executable: bool,
}

#[derive(Default, Debug, Clone, Serialize, Deserialize)]
pub struct Graph {
    pub nodes: SlotMap<NodeId, Node>,
    pub inputs: SlotMap<InputId, InputParam>,
    pub outputs: SlotMap<OutputId, OutputParam>,
    // Connects the input of a node, to the output of its predecessor that
    // produces it
    pub connections: HashMap<InputId, OutputId>,
}

pub enum InputDescriptor {
    Vector { default: Vec3<f32> },
    Mesh,
    Selection,
    Scalar { default: f32, min: f32, max: f32 },
    Enum { values: Vec<String> },
    NewFile,
}

pub struct OutputDescriptor(pub DataType);

pub struct NodeDescriptor {
    pub op_name: String,
    pub label: String,
    pub inputs: Vec<(String, InputDescriptor)>,
    pub outputs: Vec<(String, OutputDescriptor)>,
    pub is_executable: bool,
}
